#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mlonmcu.logging import get_logger

from mlonmcu.flow.tvm.backend.backend import TVMBackend
from mlonmcu.flow.tvm.backend.tvmllvm import TVMLLVMBackend

logger = get_logger()

TVM_PLATFORM_BACKEND_REGISTRY = {}


def register_tvm_platform_backend(backend_name, b, override=False):
    global TVM_PLATFORM_BACKEND_REGISTRY

    if backend_name in TVM_PLATFORM_BACKEND_REGISTRY and not override:
        raise RuntimeError(f"TVM platform backend  {backend_name} is already registered")
    TVM_PLATFORM_BACKEND_REGISTRY[backend_name] = b


def get_tvm_platform_backends():
    return TVM_PLATFORM_BACKEND_REGISTRY


register_tvm_platform_backend("tvmllvm", TVMLLVMBackend)


def create_tvm_platform_backend(name, platform, base=TVMBackend):
    class TvmPlatformBackend(base):
        def __init__(self, features=None, config=None):
            super().__init__(runtime="cpp", fmt="so", features=features, config=config)
            self.platform = platform

    return TvmPlatformBackend



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def requires_patch(context: MlonMcuContext):
    return context.environment.has_feature("disable_legalize")


def _validate_tvm(context: MlonMcuContext, params=None):
    # user_vars = context.environment.vars
    patch = bool(params.get("patch", False))
    if patch:
        if not requires_patch(context):
            return False
    return context.environment.has_framework("tvm")


def _validate_tvm_clone(context: MlonMcuContext, params=None):
    user_vars = context.environment.vars
    tvm_src_dir = user_vars.get("tvm.src_dir", None)
    if tvm_src_dir:
        return False
    return _validate_tvm(context, params=params)


def _validate_tvm_build(context: MlonMcuContext, params=None):
    user_vars = context.environment.vars
    use_tlcpack = user_vars.get("tvm.use_tlcpack", False)
    debug_build = user_vars.get("tvm.debug_build", False)
    release_build = user_vars.get("tvm.release_build", True)
    patch = bool(params.get("patch", False))
    dbg = bool(params.get("dbg", False))
    # There is not good reason to build without cmsisnn
    # cmsisnn = bool(params.get("cmsisnn", False))
    # if cmsisnn:
    #     if not (context.environment.has_feature("cmsisnnbyoc") or context.environment.has_feature("muriscvnnbyoc")):
    #         return False
    if patch:
        if not requires_patch(context):
            return False
    if use_tlcpack:
        assert not requires_patch(context)
        return False
    if dbg:
        assert not use_tlcpack
        if not debug_build:
            return False
    else:
        if not release_build:
            return False
    tvm_install_dir = user_vars.get("tvm.install_dir", None)
    if tvm_install_dir:
        return False

    return context.environment.has_framework("tvm")


def _validate_tvm_install(context: MlonMcuContext, params=None):
    return False


@Tasks.provides(["tvm.src_dir", "tvm.configs_dir", "tvm.pythonpath"])
@Tasks.optional(["tvm_extensions.src_dir"])
@Tasks.param("patch", [False, True])  # This is just a temporary workaround until the patch is hopefully upstreamed
@Tasks.validate(_validate_tvm_clone)
@Tasks.register(category=TaskType.FRAMEWORK)
def clone_tvm(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Clone the TVM repository."""
    if not params:
        params = {}
    patch = params["patch"]
    flags = utils.makeFlags((patch, "patch"))
    tvmName = utils.makeDirName("tvm", flags=flags)
    tvmSrcDir = context.environment.paths["deps"].path / "src" / tvmName
    tvmPythonPath = tvmSrcDir / "python"
    if rebuild or not utils.is_populated(tvmSrcDir):
        tvmRepo = context.environment.repos["tvm"]
        utils.clone_wrapper(tvmRepo, tvmSrcDir, refresh=rebuild)
        if patch:
            extSrcDir = context.cache["tvm_extensions.src_dir"]
            patchFile = extSrcDir / "tvmc_diff.patch"
            utils.apply(tvmSrcDir, patchFile)

    context.cache["tvm.src_dir", flags] = tvmSrcDir
    context.cache["tvm.configs_dir", flags] = tvmSrcDir / "configs"
    context.cache["tvm.pythonpath", flags] = tvmPythonPath


@Tasks.needs(["tvm.src_dir", "llvm.install_dir"])
@Tasks.provides(["tvm.build_dir", "tvm.lib"])
@Tasks.param("dbg", [False, True])
@Tasks.param("cmsisnn", [True])  # There is no good reason to build without cmsisnn
@Tasks.validate(_validate_tvm_build)
@Tasks.register(category=TaskType.FRAMEWORK)
def build_tvm(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Build the TVM framework."""
    if not params:
        params = {}
    flags_ = utils.makeFlags((params["dbg"], "dbg"), (params["cmsisnn"], "cmsisnn"))
    flags = utils.makeFlags((params["dbg"], "dbg"))
    dbg = bool(params["dbg"])
    cmsisnn = bool(params["cmsisnn"])
    # FIXME: Try to use TVM dir outside of src dir to allow multiple versions/dbg etc!
    # This should help: TVM_LIBRARY_PATH -> tvm.build_dir
    tvmName = utils.makeDirName("tvm", flags=flags)
    tvmSrcDir = context.lookup("tvm.src_dir", ())  # params["patch"] does not affect the build
    tvmBuildDir = context.environment.paths["deps"].path / "build" / tvmName
    tvmInstallDir = context.environment.paths["deps"].path / "install" / tvmName
    tvmLib = tvmBuildDir / "libtvm.so"
    user_vars = context.environment.vars
    if rebuild or not utils.is_populated(tvmBuildDir) or not tvmLib.is_file():
        ninja = False
        if "tvm.make_tool" in user_vars:
            if user_vars["tvm.make_tool"] == "ninja":
                ninja = True
        utils.mkdirs(tvmBuildDir)
        cfgFileSrc = Path(tvmSrcDir) / "cmake" / "config.cmake"
        cfgFile = tvmBuildDir / "config.cmake"
        llvmConfig = str(Path(context.cache["llvm.install_dir"]) / "bin" / "llvm-config")
        llvmConfigEscaped = str(llvmConfig).replace("/", "\\/")
        utils.copy(cfgFileSrc, cfgFile)
        utils.exec(
            "sed",
            "-i",
            "--",
            r"s/USE_MICRO OFF/USE_MICRO ON/g",
            str(cfgFile),
        )
        utils.exec(
            "sed",
            "-i",
            "--",
            r"s/USE_MICRO_STANDALONE_RUNTIME OFF/USE_MICRO_STANDALONE_RUNTIME ON/g",
            str(cfgFile),
        )
        utils.exec(
            "sed",
            "-i",
            "--",
            r"s/USE_LLVM \(OFF\|ON\)/USE_LLVM " + llvmConfigEscaped + "/g",
            str(cfgFile),
        )
        utils.exec(
            "sed",
            "-i",
            "--",
            r"s/USE_UMA OFF/USE_UMA ON/g",
            str(cfgFile),
        )
        if cmsisnn:
            utils.exec(
                "sed",
                "-i",
                "--",
                r"s/USE_CMSISNN \(OFF\|ON\)/USE_CMSISNN ON/g",
                str(cfgFile),
            )

        utils.cmake(
            tvmSrcDir,
            "-DCMAKE_INSTALL_PREFIX=" + str(tvmInstallDir),
            cwd=tvmBuildDir,
            debug=dbg,
            use_ninja=ninja,
            live=verbose,
        )
        utils.make(cwd=tvmBuildDir, threads=threads, use_ninja=ninja, live=verbose)
    context.cache["tvm.build_dir", flags] = tvmBuildDir
    context.cache["tvm.build_dir", flags_] = tvmBuildDir
    context.cache["tvm.install_dir", flags] = tvmInstallDir
    context.cache["tvm.install_dir", flags_] = tvmInstallDir
    context.cache["tvm.lib", flags] = tvmLib
    context.cache["tvm.lib", flags_] = tvmLib


@Tasks.needs(["tvm.build_dir"])
@Tasks.provides(["tvm.install_dir"])
@Tasks.param("dbg", [False, True])
@Tasks.validate(_validate_tvm_install)
@Tasks.register(category=TaskType.FRAMEWORK)
def install_tvm(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Install TVM."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "tvm.install_dir" in user_vars:
        return False
    flags = utils.makeFlags((params["dbg"], "dbg"))
    # tvmName = utils.makeDirName("tvm", flags=flags)
    # tvmSrcDir = context.cache["tvm.src_dir", ()]
    tvmBuildDir = context.cache["tvm.build_dir", flags]
    tvmInstallDir = context.cache["tvm.install_dir", flags]
    if rebuild or not utils.is_populated(tvmInstallDir):
        utils.make("install", cwd=tvmBuildDir, threads=threads, live=verbose)
    context.cache["tvm.install_dir", flags] = tvmInstallDir


def _validate_tvm_extensions(context: MlonMcuContext, params=None):
    return _validate_tvm_build(context, params=params) and requires_patch(context)


@Tasks.provides(["tvm_extensions.src_dir", "tvm_extensions.wrapper"])
@Tasks.validate(_validate_tvm_extensions)
@Tasks.register(category=TaskType.FEATURE)
def clone_tvm_extensions(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the TVM extensions repository."""
    extName = utils.makeDirName("tvm_extensions")
    extSrcDir = context.environment.paths["deps"].path / "src" / extName
    extWrapper = extSrcDir / "tvmc_wrapper.py"
    if rebuild or not utils.is_populated(extSrcDir):
        extRepo = context.environment.repos["tvm_extensions"]
        utils.clone_wrapper(extRepo, extSrcDir, refresh=rebuild)
    context.cache["tvm_extensions.src_dir"] = extSrcDir
    context.cache["tvm_extensions.wrapper"] = extWrapper



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import IntEnum

from mlonmcu.target import get_targets, Target
from mlonmcu.logging import get_logger

logger = get_logger()


def get_mlif_platform_targets():
    return get_targets()


class MlifExitCode(IntEnum):
    ERROR = 0x10
    INVALID_SIZE = 0x11
    OUTPUT_MISSMATCH = 0x12

    @classmethod
    def values(cls):
        return list(map(int, cls))


def create_mlif_platform_target(name, platform, base=Target):
    class MlifPlatformTarget(base):
        DEFAULTS = {
            **base.DEFAULTS,
        }

        def __init__(self, features=None, config=None):
            super().__init__(name=name, features=features, config=config)
            self.platform = platform
            self.validation_result = None

        def get_metrics(self, elf, directory, handle_exit=None):
            # This is wrapper around the original exec function to catch special return codes thrown by the inout data
            # feature (TODO: catch edge cases: no input data available (skipped) and no return code (real hardware))
            if self.platform.validate_outputs or not self.platform.skip_check:

                def _handle_exit(code, out=None):
                    if handle_exit is not None:
                        code = handle_exit(code, out=out)
                    if code == 0:
                        self.validation_result = True
                    else:
                        if code in MlifExitCode.values():
                            reason = MlifExitCode(code).name
                            logger.error("A platform error occured during the simulation. Reason: %s", reason)
                            self.validation_result = False
                            if not self.platform.fail_on_error:
                                code = 0
                    return code

            else:
                _handle_exit = handle_exit

            metrics, out, artifacts = super().get_metrics(elf, directory, handle_exit=_handle_exit)

            if self.platform.validate_outputs or not self.platform.skip_check:
                metrics.add("Validation", self.validation_result)
            return metrics, out, artifacts

        def get_platform_defs(self, platform):
            ret = super().get_platform_defs(platform)
            target_system = self.get_target_system()
            ret["TARGET_SYSTEM"] = target_system
            return ret

    return MlifPlatformTarget



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This file contains read lock and write lock classes based on filelock
The locks are non-blocking.
"""

from filelock import FileLock
import uuid  # this is used to create a identifier for every ReadFileLock and WriteFileLock instance.
import datetime  # this is used to track the lock actions
import os
import yaml
from pathlib import Path
import atexit


class RWLockTimeout(TimeoutError):
    """Raised when the lock could not be acquired."""

    def __init__(self, lock) -> None:
        #: The Read or Write lock instance.
        self.lock = lock

    def __str__(self) -> str:
        return f"The lock with id '{self.lock.id}' in env '{self.lock.filepath.parent}' could not be acquired."


class ReadFileLock:
    def __init__(self, filepath):
        self.filepath = Path(filepath)
        self.trackfilepath = self.filepath.parent / (str(self.filepath.stem) + "_track")
        self.lock = FileLock(self.filepath)
        self.id = str(uuid.uuid4())

    def acquire(self, raise_exception=True):
        """
        This function tried to acquire a ReadFileLock.
        The process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. check if the lock is already occupied by another write process
        4.1. release filelock and raise exception(or return 0) if the lock is already occupied by another write process
        4.2. otherwise write the updated lock occupation situation back, release filelock and return

            Parameters:
                raise_exception (bool): whether an exception should be raised when failed (default: True)

            Returns:
                success (bool): whether succeeded or not.
                    True means succeeded, False means failed (if the param raiseException is set to False).
                    A RWLockTimeout exception will be raised if failed (if the param raiseException is set to True).
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # create a file to tack the lock occupation situation
            with open(self.trackfilepath, "w") as track_file:
                pass

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. check if the lock is already occupied by another write process
        write_occupied = False
        for k, v in lock_occupy_info.items():
            if v["type"] == "write":
                write_occupied = True
                break

        # 4
        if write_occupied:
            lock_acquire_success = False
            self.lock.release()
            if raise_exception:
                raise RWLockTimeout(self)
        else:
            time_info = datetime.datetime.now().replace(microsecond=0).isoformat()
            lock_occupy_info[self.id] = {"time": f"{time_info}", "type": "read"}
            with open(self.trackfilepath, "w") as track_file:
                yaml.dump(lock_occupy_info, track_file, default_flow_style=False)
            lock_acquire_success = True

        self.lock.release()
        atexit.register(self.release)
        return lock_acquire_success

    def release(self):
        """
        This function releases a ReadFileLock.
        the process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. delete the record of self.id (no exception will be raised if self.id is not found in the record)
        4. write the updated lock occupation situation back, release filelock and return
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # raise error if file does not exist
            raise RuntimeError("track file for read write log is missing.")

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. delete the record of self.id
        if self.id in lock_occupy_info.keys():
            lock_occupy_info.pop(self.id)

        # 4. write the updated lock occupation situation back, release filelock and return
        with open(self.trackfilepath, "w") as track_file:
            if lock_occupy_info:
                yaml.dump(lock_occupy_info, track_file, default_flow_style=False)

        self.lock.release()

    @property
    def is_locked(self):
        """
        This property returns if a lock is occupied(locked) by other processes.
        the process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. check if the lock is already occupied(locked) by another write process

            Returns:
                is_locked (bool): whether the lock is already occupied(locked) by another write process
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # create a file to tack the lock occupation situation
            with open(self.trackfilepath, "w"):
                pass

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. check if the lock is already occupied by another write process
        write_occupied = False
        for k, v in lock_occupy_info.items():
            if v["type"] == "write":
                write_occupied = True
                break

        self.lock.release()
        return write_occupied


class WriteFileLock:
    def __init__(self, filepath):
        self.filepath = Path(filepath)
        self.trackfilepath = self.filepath.parent / (str(self.filepath.stem) + "_track")
        self.lock = FileLock(self.filepath)
        self.id = str(uuid.uuid4())

    def acquire(self, raise_exception=True):
        """
        This function tried to acquire a WriteFileLock.
        The process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. check if the lock is already occupied by another write process
        4.1. release filelock and raise exception(or return 0) if the lock is already occupied by another write process
        4.2. otherwise write the updated lock occupation situation back, release filelock and return

            Parameters:
                raise_exception (bool): whether raises an exception when failed (default: True)

            Returns:
                success (bool): whether succeeded or not.
                    True means succeeded, False means failed (if the param raiseException is set to False).
                    A RWLockTimeout exception will be raised if failed (if the param raiseException is set to True).
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # create a file to tack the lock occupation situation
            with open(self.trackfilepath, "w") as track_file:
                pass

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. check if the lock is already occupied by another process
        occupied = False
        for k, v in lock_occupy_info.items():
            if v["type"] == "write" or "read":
                occupied = True
                break

        # 4
        if occupied:
            lock_acquire_success = False
            self.lock.release()
            if raise_exception:
                raise RWLockTimeout(self)
        else:
            time_info = datetime.datetime.now().replace(microsecond=0).isoformat()
            lock_occupy_info[self.id] = {"time": f"{time_info}", "type": "write"}
            with open(self.trackfilepath, "w") as track_file:
                yaml.dump(lock_occupy_info, track_file, default_flow_style=False)
            lock_acquire_success = True

        self.lock.release()
        atexit.register(self.release)
        return lock_acquire_success

    def release(self):
        """
        This function releases a WriteFileLock.
        the process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. delete the record of self.id (no exception will be raised if self.id is not found in the record)
        4. write the updated lock occupation situation back, release filelock and return
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # raise error if file does not exist
            raise RuntimeError("track file for read write log is missing.")

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
                if lock_occupy_info is None:
                    lock_occupy_info = {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. delete the record of self.id
        if self.id in lock_occupy_info.keys():
            lock_occupy_info.pop(self.id)

        # 4. write the updated lock occupation situation back, release filelock and return
        with open(self.trackfilepath, "w") as track_file:
            if lock_occupy_info:
                yaml.dump(lock_occupy_info, track_file, default_flow_style=False)

        self.lock.release()

    @property
    def is_locked(self):
        """
        This property returns if a lock is occupied(locked) by other processes.
        the process is the following:
        1. acquire filelock
        2. read the lock occupation situation
        3. check if the lock is already occupied(locked) by another write process

            Returns:
                is_locked (bool): whether the lock is already occupied(locked) by another write process
        """

        # 1. acquire filelock
        self.lock.acquire(timeout=2)

        # 2. read the lock occupation situation
        if not os.path.exists(self.trackfilepath):
            # create a file to tack the lock occupation situation
            with open(self.trackfilepath, "w"):
                pass

        with open(self.trackfilepath, "r") as stream:
            # read the current lock occupation situation
            try:
                lock_occupy_info = yaml.safe_load(stream) or {}
            except yaml.YAMLError as exc:
                print(exc)

        # 3. check if the lock is already occupied by another write process
        occupied = False
        for k, v in lock_occupy_info.items():
            if v["type"] == "write" or "read":
                occupied = True
                break

        self.lock.release()
        return occupied



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import configparser
from datetime import datetime as dt

# TODO: EnvironmentList/Factory/Registry? also via code?
from .config import get_environments_file


def validate_name(name):
    # TODO: regex for valid names without spaces etc
    return True


def get_environments_map():
    result = {}
    environments_file = get_environments_file()
    if os.path.isfile(environments_file):
        parser = configparser.ConfigParser()
        parser.read(environments_file)
        for env_name in parser.sections():
            result[env_name] = parser[env_name]
    return result


def get_environment_names():
    envs_dict = get_environments_map()
    return envs_dict.keys()


def get_alternative_name(name, names):
    current = name
    i = -1
    while current in names:
        i = i + 1
        if i == 0:
            current = current + "_" + str(i)
        else:
            temp = current.split("_")
            current = "_".join(temp[:-1]) + "_" + str(i)
    return current


def register_environment(name, path, overwrite=False):
    validate_name(name)
    if not os.path.isabs(path):
        raise RuntimeError("Not an absolute path")

    environments_file = get_environments_file()
    if not os.path.isfile(environments_file):
        raise RuntimeError("Environments file does not yet exist")

    # with open(environments_file, "a") as envs_file:
    #    envs_file.write(name + "=" + path)
    config = configparser.ConfigParser()
    config.read(environments_file)
    now = dt.now()
    created = now.strftime("%Y%d%mT%H%M%S")
    if name in config.sections() and not overwrite:
        raise RuntimeError(f"Environment with name {name} does already exist")
    config[name] = {"path": path, "created": created}
    environments_file = get_environments_file()
    with open(environments_file, "w") as handle:
        config.write(handle)



# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_benchmarks(context: MlonMcuContext, params=None):
    if params:
        benchmark = params.get("benchmark", None)
        assert benchmark is not None
        if not context.environment.has_frontend(benchmark):
            return False
        user_vars = context.environment.vars
        if f"{benchmark}.src_dir" in user_vars:
            return False
    return True


# Cloning all benchmarks in a single task saves a lot of code
@Tasks.provides(["benchmarks.src_dir"])
@Tasks.param("benchmark", ["embench", "taclebench", "polybench", "mibench"])
@Tasks.validate(_validate_benchmarks)
@Tasks.register(category=TaskType.FRONTEND)
def clone_benchmarks(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the embench repo."""
    benchmark = params.get("benchmark", None)
    benchName = utils.makeDirName(benchmark)
    benchmarksSrcDir = context.environment.paths["deps"].path / "src" / "benchmarks"
    benchSrcDir = benchmarksSrcDir / benchName
    utils.mkdirs(benchmarksSrcDir)
    if rebuild or not utils.is_populated(benchSrcDir):
        benchRepo = context.environment.repos[benchmark]
        utils.clone_wrapper(benchRepo, benchSrcDir, refresh=rebuild)
    context.cache["benchmarks.src_dir"] = benchmarksSrcDir
    context.cache[f"{benchmark}.src_dir"] = benchSrcDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for managing models."""

from mlonmcu.cli.common import add_common_options, add_context_options
from mlonmcu.context.context import MlonMcuContext
import mlonmcu.models


def add_models_options(parser):
    models_parser = parser.add_argument_group("models options")
    models_parser.add_argument(
        "--detailed",
        default=False,
        action="store_true",
        help="Display more information (default: %(default)s)",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the models subcommand."""
    parser = subparsers.add_parser("models", description="Manage ML on MCU models.")
    parser.set_defaults(func=handle)
    add_common_options(parser)
    add_context_options(parser)
    add_models_options(parser)
    return parser


def handle(args):
    with MlonMcuContext(path=args.home, deps_lock="read") as context:
        mlonmcu.models.print_summary(context, detailed=args.detailed)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_llvm(context: MlonMcuContext, params=None):
    if context.environment.has_toolchain("llvm"):
        return True
    if context.environment.has_framework("tvm"):
        user_vars = context.environment.vars
        use_tlcpack = user_vars.get("tvm.use_tlcpack", False)
        if not use_tlcpack:
            return True


@Tasks.provides(["llvm.install_dir"])
@Tasks.validate(_validate_llvm)
@Tasks.register(category=TaskType.MISC)
def install_llvm(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install LLVM."""
    llvmName = utils.makeDirName("llvm")
    llvmInstallDir = context.environment.paths["deps"].path / "install" / llvmName
    user_vars = context.environment.vars
    if "llvm.install_dir" in user_vars:  # TODO: also check command line flags?
        # TODO: WARNING
        llvmInstallDir = Path(user_vars["llvm.install_dir"])
    else:
        # TODO: share helper with riscv.py
        def _helper(url):
            # candidate_exts = [".zip", ".tar"]  # .tar.xz an .tar.gz also supported
            fullUrlSplit = url.split("/")
            llvmUrl = "/".join(fullUrlSplit[:-1])
            llvmFileName, llvmFileExtension = fullUrlSplit[-1].split(".", 1)
            return llvmUrl, llvmFileName, llvmFileExtension

        if "llvm.dl_url" in user_vars:
            llvmUrl, llvmFileName, llvmFileExtension = _helper(user_vars["llvm.dl_url"])
        else:
            llvmVersion = user_vars.get("llvm.version", "14.0.0")
            llvmDist = user_vars.get("llvm.distribution", "x86_64-linux-gnu-ubuntu-18.04")
            llvmUrl = f"https://github.com/llvm/llvm-project/releases/download/llvmorg-{llvmVersion}/"
            llvmFileName = f"clang+llvm-{llvmVersion}-{llvmDist}"
            llvmFileExtension = "tar.xz"
        llvmArchive = llvmFileName + "." + llvmFileExtension
        # if rebuild or not utils.is_populated(llvmInstallDir):
        # rebuild should only be triggered if the version/url changes but we can not detect that at the moment
        if not utils.is_populated(llvmInstallDir):
            utils.download_and_extract(llvmUrl, llvmArchive, llvmInstallDir, progress=verbose)
    context.cache["llvm.install_dir"] = llvmInstallDir
    context.export_paths.add(llvmInstallDir / "bin")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import yaml


def parse_metadata(path):
    content = {}
    with open(path, "r") as yamlfile:
        try:
            content = yaml.safe_load(yamlfile)
        except yaml.YAMLError as err:  # pragma: no cover
            raise RuntimeError(f"Could not open YAML file: {path}") from err
    return content


# class ModelOperator:
#     def __init__(self, name, custom=False, registration=None):
#         self.name = name
#         self.custom = custom
#         self.registration = registration
#
#
# class ModelMetadata:
#     def __init__(
#         self,
#         author="unknown",
#         description="",
#         created_at=None,
#         references=None,
#         comment=None,
#         backend_options_map=None,
#     ):
#         self.author = author if author is not None else "unknown"
#         self.description = description if description is not None else ""
#         self.created_at = created_at
#         self.refrences = references if references is not None else []
#         self.comment = comment
#         self.backend_options_map = (
#             backend_options_map if backend_options_map is not None else {}
#         )
#
#     def __repr__(self):
#         return "ModelMetadata(" + str(vars(self)) + ")"
#
#
# class TfLitelMetadata(ModelMetadata):
#     def __init__(
#         self,
#         author="unknown",
#         description="",
#         created_at=None,
#         references=None,
#         comment=None,
#         backend_options_map=None,
#         operators=[],
#     ):
#         super().__init__(
#             author=author,
#             description=description,
#             created_at=created_at,
#             references=references,
#             comment=comment,
#             backend_options_map=backend_options_map,
#             operators=operators,
#         )
#         self.operators = operators
#
#
# def parse_metadata(path):
#     with open(path, "r") as yamlfile:
#         try:
#             content = yaml.safe_load(yamlfile)
#             if not content:
#                 # file empty
#                 return ModelMetadata()
#             if "author" in content:
#                 author = content["author"]
#             else:
#                 author = None
#             if "description" in content:
#                 description = content["description"]
#             else:
#                 description = None
#             if "created_at" in content:
#                 created_at = content["created_at"]
#             else:
#                 created_at = None
#             if "references" in content:
#                 references = content["references"]
#                 assert isinstance(references, list)
#             else:
#                 references = []
#             if "comment" in content:
#                 comment = content["comment"]
#             else:
#                 comment = None
#             backend_options_map = {}
#             if "backends" in content:
#                 backends = content["backends"]
#                 for backend in backends:
#                     backend_options = parse_model_options_for_backend(
#                         backend, backends[backend]
#                     )
#                     backend_options_map[backend] = backend_options
#             metadata = ModelMetadata(
#                 author=author,
#                 description=description,
#                 created_at=created_at,
#                 references=references,
#                 comment=comment,
#                 backend_options_map=backend_options_map,
#             )
#             return metadata
#         except yaml.YAMLError as err:
#             raise RuntimeError(f"Could not open YAML file: {path}") from err
#
#     return None



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition if the contextmanager for mlonmcu environments."""

import sys
import os
import shutil
import tempfile
from typing import List, Union
from pathlib import Path
import filelock

from mlonmcu.utils import ask_user
from mlonmcu.logging import get_logger, set_log_file
from mlonmcu.session.run import Run
from mlonmcu.session.session import Session
from mlonmcu.setup.cache import TaskCache
import mlonmcu.setup.utils as utils
from mlonmcu.plugins import process_extensions
from mlonmcu.context.read_write_filelock import ReadFileLock, WriteFileLock, RWLockTimeout

from mlonmcu.environment.environment import Environment, UserEnvironment

from mlonmcu.environment.list import get_environments_map
from mlonmcu.environment.config import get_environments_dir, get_plugins_dir

logger = get_logger()


def lookup_environment() -> Environment:
    """Helper function to automatically find a suitable environment.

    This function is used if neither a name nor a path of the environment was specified by the user.
    The lookup follows a predefined order:
    - Check current working directory
    - Check MLONMCU_HOME environment variable
    - Default environment for current user

    Returns
    -------
    environment : Path
        The environment (if the lookup was successful).
    """
    logger.debug("Starting lookup for mlonmcu environment")

    logger.debug("First checking in local working directory")
    path = os.path.join(os.getcwd(), "environment.yml")
    if os.path.exists(path):
        logger.debug("Found environment directory: %s", path)
        return path

    logger.debug("Next checking environment variables")
    home = os.environ.get("MLONMCU_HOME")
    if home:
        path = os.path.join(home, "environment.yml")
        if os.path.exists(path):
            logger.debug("Found environment directory: %s", path)
            return path

    logger.debug("Looking for default environment for current user")
    envs_list = get_environments_map()
    if "default" in envs_list:
        assert "path" in envs_list["default"]
        directory = envs_list["default"]["path"]
        path = os.path.join(directory, "environment.yml")
        if os.path.exists(path):
            logger.debug("Found environment directory: %s", path)
            return path
    return None


def get_environment_by_path(path: Union[str, Path]) -> Environment:
    """Utility to find an environment file using a supplied path.

    Parameters
    ----------
    path : str/Path
        The path of the environment (or its YAML file).

    Returns
    -------
    Environment:
        The environment (if the lookup was successful).
    """
    if isinstance(path, str):
        path = Path(path)
    assert isinstance(path, Path)
    if path.is_dir():
        path = path / "environment.yml"
    if path.is_file():
        return path
    return None


def get_environment_by_name(name: str) -> Environment:
    """Utility to find an environment file using a supplied name.

    Parameters
    ----------
    name : str
        The name/alias if the environment.

    Returns
    -------
    Environment :
        The environment (if the lookup was successful).
    """
    # TODO: parse the ini file instead
    environments_dir = get_environments_dir()
    if environments_dir.is_dir():
        path = environments_dir / name
        if path.is_dir():
            return get_environment_by_path(path)
    return None


def get_ids(directory: Path) -> List[int]:
    """Get a sorted list of ids for sessions/runs found in the given directory.

    Parameters
    ----------
    directory : Path
        Directory where the sessions/runs are stored.

    Returns:
    list
        List of integers representing the session numbers. Empty list if directory does not exist.
    """
    if not directory.is_dir():
        return []

    ids = [int(o) for o in os.listdir(directory) if os.path.isdir(directory / o) and not os.path.islink(directory / o)]
    return sorted(ids)  # TODO: sort by session datetime?


def load_recent_sessions(env: Environment, count: int = None) -> List[Session]:
    """Get a list of recent sessions for the environment.

    Parameters
    ----------
    env : Environment
        MLonMCU environment which should be used.
    count : int
        Maximum number of sessions to return. Collect all if None.

    Returns
    -------
    list:
        The resulting list of session objects.
    """

    if count is not None:
        raise NotImplementedError()
    sessions = []

    sessions_directory = env.paths["temp"].path / "sessions"

    # TODO: in the future also strs (custom or hash) should be allowed
    session_ids = get_ids(sessions_directory)

    for sid in session_ids:
        session_directory = sessions_directory / str(sid)
        # session_file = sessions_directory / str(sid) / "session.txt"
        # if not session_file.is_file():
        #     continue
        runs_directory = session_directory / "runs"
        run_ids = get_ids(runs_directory)
        runs = []
        for rid in run_ids:
            run_directory = runs_directory / str(rid)
            # run_file = run_directory / "run.txt"
            # run = Run.from_file(run_file)  # TODO: actually implement run restore
            run = Run()  # TODO: fix
            run.archived = True
            run.dir = run_directory
            runs.append(run)
        session = Session(idx=sid, archived=True, dir=session_directory)
        session.runs = runs
        session.dir = session_directory
        sessions.append(session)
    return sessions


def resolve_environment_file(name: str = None, path: str = None) -> Path:
    """Utility to find the environment file by a optionally given name or path.

    The lookup is performed in a predefined order:
    - If specified: name/path
    - Else: see lookup_environment()

    Parameters
    ----------
    name : str
        Hint for the environment name provided by the user.
    path : str
        Hint for the environment path provided by the user.

    Returns
    -------
    Path :
        Path to the found environment.yml (if sucessful)
    """
    if name and path:
        raise RuntimeError("mlonmcu environments are specified either by name OR path")
    if name:
        env_file = get_environment_by_name(name)
    elif path:
        env_file = get_environment_by_path(path)
    else:
        env_file = lookup_environment()
        if not env_file:
            raise RuntimeError("Lookup for mlonmcu environment was not successful.")
    return Path(env_file)


def setup_logging(environment):
    """Check logging settings for environment and initialize the logs directory.

    Attributes
    ----------
    environment : Environment
        The MLonMCU Environment where paths, repos, features,... are configured.
    """
    defaults = environment.defaults
    level = defaults.log_level
    to_file = defaults.log_to_file
    rotate = defaults.log_rotate
    if to_file:
        assert "logs" in environment.paths, "To use a logfile, define a logging directory in your environment.yml"
        directory = environment.paths["logs"].path
        if not directory.is_dir():
            directory.mkdir()
        path = directory / "mlonmcu.log"
        set_log_file(path, level=level, rotate=rotate)


class MlonMcuContext:
    """Contextmanager for mlonmcu environments.

    Attributes
    ----------
    environment : Environment
        The MLonMCU Environment where paths, repos, features,... are configured.
    deps_lock : str ("read" or "write" default "write")
        Read means that the program does not write to the ./deps folder in the env folder.
    sessions : list
        List of sessions for the current environment.
    session_idx : list
        A counter for determining the next session index.
    cache : TaskCache
        The cache where paths of installed dependencies can be looked up.


    """

    def __init__(self, name: str = None, path: str = None, deps_lock: str = "write"):
        env_file = resolve_environment_file(name=name, path=path)
        assert env_file is not None, "Unable to find a MLonMCU environment"
        self.environment = UserEnvironment.from_file(env_file)  # TODO: move to __enter__
        actual_home = env_file.parent.resolve()
        specified_home = Path(self.environment.home).resolve()
        if specified_home != actual_home:
            logger.warning(
                "The HOME directory %s specified in environment.yml does not match workspace directory %s!",
                specified_home,
                actual_home,
            )
        setup_logging(self.environment)
        assert deps_lock in ["read", "write"]
        if deps_lock == "read":
            self.deps_lock = ReadFileLock(os.path.join(self.environment.home, ".deps_lock"))
        elif deps_lock == "write":
            self.deps_lock = WriteFileLock(os.path.join(self.environment.home, ".deps_lock"))
        self.latest_session_link_lock = filelock.FileLock(
            os.path.join(self.environment.home, ".latest_session_link_lock")
        )
        # Reusing lock for latest session link here...
        try:
            lock = self.latest_session_link_lock.acquire(timeout=10)
        except filelock.Timeout as err:
            raise RuntimeError("Lock on current context could not be aquired.") from err
        else:
            with lock:
                self.sessions = load_recent_sessions(self.environment)
                if self.environment.defaults.cleanup_auto:
                    logger.debug("Cleaning up old sessions automaticaly")
                    self.cleanup_sessions(keep=self.environment.defaults.cleanup_keep, interactive=False)
                    self.sessions = load_recent_sessions(self.environment)
                self.session_idx = self.sessions[-1].idx if len(self.sessions) > 0 else -1
                logger.debug(f"Restored {len(self.sessions)} recent sessions")
        self.cache = TaskCache()
        self.export_paths = set()

    def create_session(self, label="", config=None):
        try:
            lock = self.latest_session_link_lock.acquire(timeout=10)
        except filelock.Timeout as err:
            raise RuntimeError("Lock on current context could not be aquired.") from err
        else:
            with lock:
                """Create a new session in the current context."""
                self.sessions = load_recent_sessions(self.environment)
                idx = self.session_idx + 1
                logger.debug("Creating a new session with idx %s", idx)
                temp_directory = self.environment.paths["temp"].path
                sessions_directory = temp_directory / "sessions"
                sessions_directory.mkdir(exist_ok=True, parents=True)
                session_dir = sessions_directory / str(idx)
                session = Session(idx=idx, label=label, dir=session_dir, config=config)
                self.sessions.append(session)
                self.session_idx = idx
                # TODO: move this to a helper function
                session_link = sessions_directory / "latest"
                if os.path.islink(session_link):
                    os.unlink(session_link)
                os.symlink(session_dir, session_link)
        return session

    def load_cache(self):
        """If available load the cache.ini file in the deps directory"""
        if self.environment:
            if self.environment.paths:
                if "deps" in self.environment.paths:
                    deps_dir = self.environment.paths["deps"].path
                    if deps_dir.is_dir():
                        cache_file = deps_dir / "cache.ini"
                        if cache_file.is_file():
                            logger.info("Loading environment cache from file")
                            self.cache.read_from_file(cache_file)
                            logger.info("Successfully initialized cache")
                            return
        logger.info("No cache found in deps directory")

    def load_extensions(self):
        """If available load the extensions.py scripts in the plugin directories"""

        # TODO: check vars.enable_extensions before!
        def _load(plugins_dir, hint="Unknown"):
            if plugins_dir.is_dir():
                extensions_file = plugins_dir / "extensions.py"
                if extensions_file.is_file():
                    logger.info(f"Loading extensions.py ({hint})")
                    process_extensions(extensions_file)

        # global (user)
        plugins_dir = Path(get_plugins_dir())
        _load(plugins_dir, hint="User")

        # local (environment)
        if self.environment:
            if self.environment.paths:
                if "plugins" in self.environment.paths:
                    plugins_dir = self.environment.paths["plugins"].path
                    _load(plugins_dir, hint="Environment")

    def get_session(self, label="", resume=False, config=None) -> Session:
        """Get an active session if available, else create a new one.

        Returns
        -------
        Session:
            An active session
        """
        if resume:
            assert len(self.sessions) > 0, "There is no recent session available"
            # session = self.sessions[-1]
            assert False, "The latest session can not be resumed"
            raise NotImplementedError

        if self.session_idx < 0 or not self.sessions[-1].active:
            self.create_session(label=label, config=config)
        return self.sessions[-1]

    def __enter__(self):
        logger.debug("Enter MlonMcuContext")
        if self.deps_lock.is_locked:
            raise RuntimeError(
                "Lock on current context could not be acquired. "
                f"Current context is locked via: {self.deps_lock.filepath}"
            )
        if self.deps_lock:
            logger.debug("Locking context")
            try:
                self.deps_lock.acquire()
            except RWLockTimeout as err:
                raise RuntimeError(
                    "Lock on current context could not be aquired. "
                    f"Current context is locked via: {self.deps_lock.filepath}"
                ) from err
        self.load_cache()
        self.load_extensions()
        return self

    def cleanup(self):
        """Clean up the context before leaving the context by closing all active sessions"""
        logger.debug("Cleaning up active sessions")
        for session in self.sessions:
            if session.active:
                session.close()

    @property
    def is_clean(self):
        """Return true if all sessions in the context are inactive"""
        return not any(sess.active for sess in self.sessions)

    # WARNING: this will remove the actual session directories!
    def cleanup_sessions(self, keep=10, interactive=True):
        """Utility to cleanup old sessions from the disk."""
        assert self.is_clean
        all_sessions = self.sessions
        to_keep = all_sessions[-keep:] if keep > 0 else []
        to_remove = self.sessions[:-keep] if keep > 0 else self.sessions
        count = len(to_remove)
        if count > 0:
            temp_dir = self.environment.lookup_path("temp").path
            sessions_dir = temp_dir / "sessions"
            if interactive:
                print(
                    f"The following {count} sessions will be removed from the environments temp directory ({temp_dir}):"
                )
                print(" ".join([str(session.idx) for session in to_remove]))

            if ask_user("Are your sure?", default=not interactive, interactive=interactive):
                for session in to_remove:
                    session_dir = sessions_dir / str(session.idx)
                    if not session_dir.is_dir():
                        # Skip / Dir does not exist
                        continue
                    session_lock = session_dir / ".lock"
                    if session_lock.is_file():
                        # Skip / Session locked (unclean or in progress)
                        continue
                    shutil.rmtree(session_dir)
                self.sessions = to_keep
                self.session_idx = self.sessions[-1].idx if len(self.sessions) > 0 else -1
                if interactive:
                    print("Done")
            else:
                if interactive:
                    print("Aborted")
        else:
            if interactive:
                print("No sessions selected for removal")
        # We currently do not support rewirting the indices to start from scratch again as this
        # would lead to inconsitencies with the path in the report/cmake build dirtectory

    def get_sessions_runs_idx(self):
        sessions_dict = {}
        for session in self.sessions:
            session_runs_idx = [run.idx if run.idx is not None else i for i, run in enumerate(session.runs)]
            sessions_dict[session.idx] = session_runs_idx
        return sessions_dict

    def print_summary(self, sessions=True, runs=False, labels=True):
        def print_sessions(sessions_runs, with_runs=False, with_labels=True):
            print("Sessions:")
            session_index_map = {session.idx: i for i, session in enumerate(self.sessions)}
            for session, runs in sessions_runs.items():
                print(f"  - {session}", end="")
                if with_runs:
                    print(" (" + self.sessions[session_index_map[session]].label + ")", end="")
                if with_runs:
                    print(" [runs: " + " ".join([str(run) for run in runs]) + "]", end="")
                print()

        print("Context Summary\n")
        sessions_runs = self.get_sessions_runs_idx()
        print_sessions(sessions_runs, with_runs=runs, with_labels=labels)

    def lookup(self, key, flags=None):
        user_vars = self.environment.vars
        if key in user_vars:
            return user_vars[key]
        return self.cache[key, flags]

    def export(self, dest, session_ids=None, run_ids=None, interactive=True):
        dest = Path(dest)
        if (dest.is_file() and dest.exists()) or (dest.is_dir() and utils.is_populated(dest)):
            if not ask_user("Destination is already populated! Overwrite?", default=True, interactive=interactive):
                print("Aborted")
                return
        dest_, ext = os.path.splitext(dest)
        if session_ids is None:
            # Can not select all sessions, fall back to latest session
            session_ids = [-1]

        if run_ids is not None:
            assert len(session_ids) == 1, "Can only choose runs of a single session"

        def find_session(sid):
            if len(self.sessions) == 0:
                return None

            if sid == -1:
                assert len(self.sessions) > 0
                return self.sessions[-1]

            for session in self.sessions:
                if session.idx == sid:
                    return session
            return None

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpdir = Path(tmpdirname)
            for sid in session_ids:
                session = find_session(sid)
                if session is None:
                    print(
                        f"Lookup for session id {sid} failed. Available:", " ".join([str(s.idx) for s in self.sessions])
                    )
                    sys.exit(1)
                if len(session_ids) == 1:
                    base = tmpdir
                else:
                    base = tmpdir / str(sid)
                if run_ids is None:
                    src = session.dir / "runs"
                    report_path = session.dir / "report.csv"  # TODO: only works for csv
                    # TODO: use artifacts from restored session instead
                    shutil.copytree(
                        src, base, dirs_exist_ok=True, symlinks=True
                    )  # Warning: dirs_exist_ok=True requires python 3.8+
                    if report_path.is_file():
                        shutil.copyfile(report_path, base / "report.csv")
                else:
                    base = base / "runs"
                    for rid in run_ids:
                        if rid >= len(session.runs):
                            print(
                                f"Lookup for run id {rid} failed in session {sid}. Available:",
                                " ".join([str(i) for i in range(len(session.runs))]),
                            )
                            sys.exit(1)
                        run = session.runs[rid]  # TODO: We currently do not check if the index actually exists
                        if len(run_ids) == 1 and len(session_ids) == 1:
                            run_base = tmpdir
                        else:
                            run_base = base / str(rid)
                        src = run.dir
                        shutil.copytree(
                            src, run_base, dirs_exist_ok=True
                        )  # Warning: dirs_exist_ok=True requires python 3.8+
            if ext in [".zip", ".tar"]:
                print(f"Creating archive: {dest}")
                shutil.make_archive(dest_, ext[1:], tmpdirname)
            else:
                print(f"Creating directory: {dest}")
                if dest.is_dir():
                    shutil.rmtree(dest)  # Cleanup old contents
                # dest.mkdir(exist_ok=True)
                shutil.move(tmpdirname, str(dest))
        print("Done")

    def __exit__(self, exception_type, exception_value, traceback):
        logger.debug("Exit MlonMcuContext")
        self.cleanup()
        if self.deps_lock:
            logger.debug("Releasing lock on context")
            self.deps_lock.release()
        return False



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import mock
from mlonmcu.setup.task import get_combs, TaskFactory, TaskType, TaskGraph
from mlonmcu.setup.setup import Setup

TestTaskFactory = TaskFactory()


def _validate_example_task1(context, params={}):
    print("_validate_example_task1", params)
    assert "foo" in params and "bar" in params and "special" in params
    if params["foo"] == 0 or params["bar"] == "B":
        return params["special"]
    return False


@TestTaskFactory.provides(["dep0"])
@TestTaskFactory.param("special", [True, False])
@TestTaskFactory.param("foo", [0, 1])
@TestTaskFactory.param("bar", ["A", "B"])
@TestTaskFactory.validate(_validate_example_task1)
@TestTaskFactory.register(category=TaskType.MISC)
def example_task1(context, params={}, rebuild=False):
    context.cache._vars["dep0"] = ""
    pass


@TestTaskFactory.needs(["dep0"])
@TestTaskFactory.provides(["dep1"])
@TestTaskFactory.register(category=TaskType.MISC)
def example_task2(context, params={}, rebuild=False):
    context.cache._vars["dep1"] = ""
    pass


def test_task_registry():
    assert len(TestTaskFactory.registry) == 2
    assert len(TestTaskFactory.validates) == 1
    assert len(TestTaskFactory.dependencies) == 1
    assert len(TestTaskFactory.providers) == 2
    assert len(TestTaskFactory.types) == 2
    assert len(TestTaskFactory.params) == 2
    assert len(TestTaskFactory.params["example_task1"]) == 3


def test_task_registry_reset_changes():
    TestTaskFactory.changed = ["dep0"]
    TestTaskFactory.reset_changes()
    assert len(TestTaskFactory.changed) == 0


@pytest.mark.parametrize("progress", [False, True])
@pytest.mark.parametrize("print_output", [False, True])
@pytest.mark.parametrize("rebuild", [False, True])  # TODO: actually test this
@pytest.mark.parametrize("write_cache", [False])  # TODO: True
@pytest.mark.parametrize("write_env", [False])  # TODO: True
def test_setup_install_dependencies(progress, print_output, rebuild, write_cache, write_env, fake_context):
    # example_task1_mock = mock.Mock(return_value=True)
    TestTaskFactory.registry["example_task1"] = mock.Mock(return_value=True)
    # example_task2_mock = mock.Mock(return_value=True)
    TestTaskFactory.registry["example_task2"] = mock.Mock(return_value=True)
    config = {"print_output": print_output}
    installer = Setup(config=config, context=fake_context, tasks_factory=TestTaskFactory)
    result = installer.install_dependencies(
        progress=progress, write_cache=write_cache, write_env=write_env, rebuild=rebuild
    )
    assert result
    # assert example_task1_mock.call_count == 3
    assert (
        TestTaskFactory.registry["example_task1"].call_count == 1
    )  # Due to the mock, the actual wrapper is not executed anymore, params are not considered etc
    # assert example_task2_mock.call_count == 1
    assert TestTaskFactory.registry["example_task2"].call_count == 1


def test_task_get_combs():
    assert get_combs({}) == []
    assert get_combs({"foo": []}) == []  # TODO: invalid?
    assert get_combs({"foo": [0]}) == [{"foo": 0}]
    # assert set(get_combs({"foo": [0, 1, 2]})) == set([{"foo": 0}, {"foo": 1}, {"foo": 2}])
    assert get_combs({"foo": [0, 1, 2]}) == [{"foo": 0}, {"foo": 1}, {"foo": 2}]
    assert len(get_combs({"foo": [0, 1], "bar": ["A", "B"]})) == 4
    assert len(get_combs({"foo": [0, 1], "bar": ["A", "B"]})[0].items()) == 2


def test_task_graph():
    names = ["NodeA", "NodeB", "NodeC"]
    dependencies = {"NodeA": [], "NodeB": ["foo", "bar"], "NodeC": ["foo"]}
    providers = {"foo": "NodeA", "bar": "NodeC"}
    task_graph = TaskGraph(names, dependencies, providers)
    nodes, edges = task_graph.get_graph()
    print("nodes", nodes)
    print("edges", edges)
    assert len(nodes) == len(names)
    assert len(edges) == sum([len(deps) for deps in dependencies.values()])
    order = task_graph.get_order()
    assert len(order) == len(nodes)
    assert order.index("NodeB") > order.index("NodeA") and order.index("NodeB") > order.index("NodeC")
    assert order.index("NodeC") > order.index("NodeA")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import _validate_gcc, get_task_factory
from .corstone300 import _validate_corstone300

logger = get_logger()
Tasks = get_task_factory()


def _validate_arm_gcc(context: MlonMcuContext, params=None):
    return _validate_corstone300(context, params=params) and _validate_gcc(context, params=params)


@Tasks.provides(["arm_gcc.install_dir"])
@Tasks.validate(_validate_corstone300)
@Tasks.register(category=TaskType.TARGET)
def install_arm_gcc(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install GNU compiler toolchain from ARM."""
    armName = utils.makeDirName("arm_gcc")
    armInstallDir = context.environment.paths["deps"].path / "install" / armName
    user_vars = context.environment.vars
    if "arm_gcc.install_dir" in user_vars:  # TODO: also check command line flags?
        # armInstallDir = user_vars["riscv_gcc.install_dir"]
        return False
    else:
        if not utils.is_populated(armInstallDir):
            armUrl = "https://developer.arm.com/-/media/Files/downloads/gnu/11.2-2022.02/binrel/"
            armFileName = "gcc-arm-11.2-2022.02-x86_64-arm-none-eabi"
            armArchive = armFileName + ".tar.xz"
            utils.download_and_extract(armUrl, armArchive, armInstallDir, progress=verbose)
    context.cache["arm_gcc.install_dir"] = armInstallDir
    context.export_paths.add(armInstallDir / "bin")



import pytest

# from mlonmcu.session.run import RunStage

# from mlonmcu.artifact import ArtifactFormat
from mlonmcu.artifact import lookup_artifacts

from mlonmcu.testing.helpers import _test_frontend

# from mlonmcu.testing.helpers import _check_features, _init_run


# TODO: make sure that we use quant/float models and several different operators
DEFAULT_TFLITE_MODELS = ["sine_model"]
DEFAULT_ONNX_MODELS = ["onnx_mnist"]
DEFAULT_RELAY_MODELS = ["test_cnn"]
DEFAULT_PB_MODELS = ["mobilenet_v1_1_0_224"]
ALL_PB_MODELS = ["mobilenet_v1_1_0_224", "mobilenet_v1_1_0_224_quant", "mobilenet_v1_0_25_128"]
DEFAULT_PADDLE_MODELS = ["paddle_resnet50"]


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
@pytest.mark.parametrize("feature_names", [[], ["tflite_visualize"]])
@pytest.mark.parametrize("config", [{}])
def test_frontend_tflite(user_context, model_name, models_dir, feature_names, config):
    _, artifacts = _test_frontend("tflite", user_context, model_name, models_dir, feature_names, config)


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_TFLITE_MODELS)
# @pytest.mark.parametrize("feature_names", [[], ["tflite_visualize"]])
@pytest.mark.parametrize("feature_names", [["visualize"]])  # TODO: rename
@pytest.mark.parametrize("config", [{}])
def test_feature_tflite_visualize(user_context, model_name, models_dir, feature_names, config):
    _, artifacts = _test_frontend("tflite", user_context, model_name, models_dir, feature_names, config)
    assert len(lookup_artifacts(artifacts, name="tflite_visualize.html")) == 1


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_ONNX_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_frontend_onnx(user_context, model_name, models_dir, feature_names, config):
    _test_frontend("onnx", user_context, model_name, models_dir, feature_names, config)


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_RELAY_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_frontend_relay(user_context, model_name, models_dir, feature_names, config):
    _test_frontend("relay", user_context, model_name, models_dir, feature_names, config)


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_RELAY_MODELS)
@pytest.mark.parametrize("feature_names", [["relayviz"]])
@pytest.mark.parametrize("config", [{"relayviz.plotter": "term"}, {"relayviz.plotter": "dot"}])
def test_feature_relayviz(user_context, model_name, models_dir, feature_names, config):
    _, artifacts = _test_frontend("relay", user_context, model_name, models_dir, feature_names, config)
    ext = "txt" if config["relayviz.plotter"] == "term" else "pdf"
    assert len(lookup_artifacts(artifacts, name=f"relayviz.{ext}")) == 1


@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_PB_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_frontend_pb(user_context, model_name, models_dir, feature_names, config):
    _test_frontend("pb", user_context, model_name, models_dir, feature_names, config)


@pytest.mark.skip(reason="unimplemented")
@pytest.mark.user_context
@pytest.mark.parametrize("model_name", DEFAULT_PADDLE_MODELS)
@pytest.mark.parametrize("feature_names", [[]])
@pytest.mark.parametrize("config", [{}])
def test_frontend_paddle(user_context, model_name, models_dir, feature_names, config):
    _test_frontend("paddle", user_context, model_name, models_dir, feature_names, config)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" MLonMCU Features API"""

from abc import ABC

from mlonmcu.config import filter_config, str2bool
from .type import FeatureType


# TODO: features might get an optional context parameter to lookup if they are supported by themselfs in the environment


class FeatureBase(ABC):
    """Feature base class"""

    feature_type = None
    scope = None

    DEFAULTS = {"enabled": True}
    REQUIRED = set()
    OPTIONAL = set()

    def __init__(self, name, features=None, config=None):
        self.name = name
        self.used = False
        self.config = config if config else {}
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        # assert features is None, "Features with features are currently not supported"
        # just ignore the supplied feature list for now

    @property
    def enabled(self):
        value = self.config["enabled"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def remove_config_prefix(self, config):  # TODO: move to different place
        def helper(key):
            return key.split(f"{self.name}.")[-1]

        return {helper(key): value for key, value in config.items() if f"{self.name}." in key}

    def __repr__(self):
        return type(self).__name__ + f"({self.name})"

    # @property
    # def types(self):
    #     return [base.feature_type for base in type(self).__bases__]

    @classmethod
    def types(cls):
        """Find out which types the features is based on."""
        return [base.feature_type for base in cls.__bases__]

    # This does not make sense because the get_?_config methods may beed a parameter
    # This could be solved by seeting he backend/target/frontend in the constructor!
    # Multiple inheritance would make this still pretty dirty
    # def get_config(self):
    #     for feature_type in self.types:
    #         type_name = FeatureType(feature_type).name.lower()
    #         method_name = f"get_{type_name}_config"
    #         method = getattr(self, method_name)
    #         args = {"type_name": getattr(self, type_name)}
    #         self.method(**args)


class Feature(FeatureBase):
    """Feature of unknown type"""

    feature_type = FeatureType.OTHER


class FrontendFeature(FeatureBase):
    """Frontend related feature"""

    feature_type = FeatureType.FRONTEND

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_frontend_config(self, frontend):
        # pylint: disable=unused-argument
        return {}

    def add_frontend_config(self, frontend, config):
        config.update(self.get_frontend_config(frontend))

    def update_formats(self, frontend, input_formats, output_formats):
        pass


class FrameworkFeature(FeatureBase):
    """Framework related feature"""

    feature_type = FeatureType.FRAMEWORK

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_framework_config(self, framework):
        # pylint: disable=unused-argument
        return {}

    def add_framework_config(self, framework, config):
        config.update(self.get_framework_config(framework))


class BackendFeature(FeatureBase):
    """Backend related feature"""

    feature_type = FeatureType.BACKEND

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_backend_config(self, backend):
        # pylint: disable=unused-argument
        return {}

    def add_backend_config(self, backend, config):
        # TODO: cfg passed to method instead of contructor or self.config = config
        config.update(self.get_backend_config(backend))


class TargetFeature(FeatureBase):
    """Target related feature"""

    feature_type = FeatureType.TARGET

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_target_config(self, target):
        # pylint: disable=unused-argument
        return {}

    def add_target_config(self, target, config):
        # TODO: cfg passed to method instead of contructor or self.config = config
        config.update(self.get_target_config(target))

    def get_target_callbacks(self, target):
        return None, None

    def add_target_callbacks(self, target, pre_callbacks, post_callbacks):
        # TODO: cfg passed to method instead of contructor or self.config = config
        pre_callback, post_callback = self.get_target_callbacks(target)
        if pre_callback is not None and pre_callbacks is not None:
            pre_callbacks.append(pre_callback)
        if post_callback is not None and post_callbacks is not None:
            post_callbacks.append(post_callback)


class PlatformFeature(FeatureBase):
    """Platform/Compile related feature"""

    feature_type = FeatureType.PLATFORM

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        return {}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))


class SetupFeature(FeatureBase):  # TODO: alternative: CacheFeature
    """Setup/Cache related feature"""

    feature_type = FeatureType.SETUP

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_setup_config(self):
        raise NotImplementedError
        return {}

    def add_setup_config(self, config):
        raise NotImplementedError
        config.update(self.get_setup_config(compile))

    def get_required_cache_flags(self):
        return {}

    def add_required_cache_flags(self, required_flags):
        own_flags = self.get_required_cache_flags()
        for key, flags in own_flags.items():
            if key in required_flags:
                # remove duplicates along the way
                required_flags[key] = list(set(required_flags[key] + flags))
            else:
                required_flags[key] = flags


class RunFeature(FeatureBase):
    """Run related feature"""

    feature_type = FeatureType.RUN

    def __init__(self, name, features=None, config=None):
        super().__init__(name, features=features, config=config)

    def get_run_config(self):
        return {}

    def add_run_config(self, config):
        config.update(self.get_run_config())



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_dtc(context: MlonMcuContext, params=None):
    # TODO: require this for: ???
    # TODO: cleanup after build
    # return True
    return False


@Tasks.provides(["dtc.src_dir"])
@Tasks.validate(_validate_dtc)
@Tasks.register(category=TaskType.MISC)
def clone_dtc(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Clone the dtc repo."""
    name = utils.makeDirName("dtc")
    srcDir = context.environment.paths["deps"].path / "src" / name
    user_vars = context.environment.vars
    if "dtc.src_dir" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(srcDir):
        repo = context.environment.repos["dtc"]
        utils.clone(repo.url, srcDir, branch=repo.ref)
    context.cache["dtc.src_dir"] = srcDir


@Tasks.needs(["dtc.src_dir"])
@Tasks.provides(["dtc.install_dir", "dtc.build_dir", "dtc.exe"])
@Tasks.validate(_validate_dtc)
@Tasks.register(category=TaskType.MISC)
def build_dtc(context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()):
    """Build the device tree compile."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    if "dtc.exe" in user_vars:  # TODO: also check command line flags?
        return False
    name = utils.makeDirName("dtc")
    srcDir = context.cache["dtc.src_dir"]
    buildDir = context.environment.paths["deps"].path / "build" / name
    installDir = context.environment.paths["deps"].path / "install" / name
    exe = installDir / "bin" / "dtc"
    if rebuild or not (utils.is_populated(buildDir) and exe.is_file()):
        utils.mkdirs(buildDir)
        # utils.make("install", "-C", srcDir, f"PREFIX={installDir}", cwd=buildDir, threads=threads, live=verbose)
        utils.make("dtc", "-C", srcDir, f"PREFIX={installDir}", cwd=buildDir, threads=threads, live=verbose)
        utils.mkdirs(installDir / "bin")
        utils.copy(srcDir / "dtc", installDir / "bin" / "dtc")
    context.cache["dtc.build_dir"] = buildDir
    context.cache["dtc.install_dir"] = installDir
    context.cache["dtc.exe"] = exe
    context.export_paths.add(installDir)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_cmsisnn(context: MlonMcuContext, params=None):
    if not (context.environment.has_feature("cmsisnn") or context.environment.has_feature("cmsisnnbyoc")):
        return False
    mvei = params.get("mvei", False)
    dsp = params.get("dsp", False)
    target_arch = params.get("target_arch", None)
    if target_arch == "arm":
        if dsp and not context.environment.has_feature("arm_dsp"):
            return False
        if mvei and not context.environment.has_feature("arm_mvei"):
            return False
    else:
        if mvei or dsp:
            return False
    return True


@Tasks.provides(["cmsisnn.dir"])
@Tasks.validate(_validate_cmsisnn)
@Tasks.register(category=TaskType.OPT)
def clone_cmsisnn(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """CMSIS-NN repository."""
    cmsisName = utils.makeDirName("cmsisnn")
    cmsisnnSrcDir = Path(context.environment.paths["deps"].path) / "src" / cmsisName
    if rebuild or not utils.is_populated(cmsisnnSrcDir):
        cmsisnnRepo = context.environment.repos["cmsisnn"]
        utils.clone_wrapper(cmsisnnRepo, cmsisnnSrcDir, refresh=rebuild)
    context.cache["cmsisnn.dir"] = cmsisnnSrcDir


def _validate_cmsis(context: MlonMcuContext, params=None):
    return _validate_cmsisnn(context, params=params) or context.environment.has_target("corstone300")


@Tasks.provides(["cmsis.dir"])
@Tasks.validate(_validate_cmsis)
@Tasks.register(category=TaskType.MISC)
def clone_cmsis(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """CMSIS repository."""
    cmsisName = utils.makeDirName("cmsis")
    cmsisSrcDir = Path(context.environment.paths["deps"].path) / "src" / cmsisName
    if rebuild or not utils.is_populated(cmsisSrcDir):
        cmsisRepo = context.environment.repos["cmsis"]
        utils.clone_wrapper(cmsisRepo, cmsisSrcDir, refresh=rebuild)
    context.cache["cmsis.dir"] = cmsisSrcDir


def _validate_ethosu_platform(context: MlonMcuContext, params=None):
    return context.environment.has_target("corstone300")


@Tasks.provides(["ethosu_platform.dir"])
@Tasks.validate(_validate_ethosu_platform)
@Tasks.register(category=TaskType.MISC)
def clone_ethosu_platform(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """EthosU platform repository."""
    ethosuPlatformName = utils.makeDirName("ethosu_platform")
    ethosuPlatformSrcDir = Path(context.environment.paths["deps"].path) / "src" / ethosuPlatformName
    if rebuild or not utils.is_populated(ethosuPlatformSrcDir):
        ethosuPlatformRepo = context.environment.repos["ethosu_platform"]
        utils.clone_wrapper(ethosuPlatformRepo, ethosuPlatformSrcDir, refresh=rebuild)
    context.cache["ethosu_platform.dir"] = ethosuPlatformSrcDir



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions for benchmarking."""
import re
import ast


def parse_bench_results(out, allow_missing=False, target_name=None):
    lines = out.split("\n")
    for i, line in enumerate(lines):
        if "Program start." in line:
            lines = lines[i:]
            break
    for i, line in enumerate(lines):
        search = "Program finish."
        if target_name and target_name == "ara_rtl":
            search = "gram finish."
        if search in line:
            lines = lines[: i + 1]
            break
    out = "\n".join(lines)
    matches = re.compile(r"# (.+): ([0-9.,E-]+)").findall(out)
    ret = {}
    for key, value in matches:
        value = ast.literal_eval(value)
        assert isinstance(value, (int, float))
        ret[key] = value
    for mode in ["Setup", "Run", "Total"]:
        cycles = ret.get(f"{mode} Cycles", None)
        instructions = ret.get(f"{mode} Instructions", None)
        if cycles is None or instructions is None:
            continue
        if cycles == 0 or instructions == 0:
            continue
        ret[f"{mode} CPI"] = cycles / instructions
    return ret


def add_bench_metrics(out, metrics, allow_missing=False, target_name=None):
    results = parse_bench_results(out, allow_missing=allow_missing, target_name=target_name)
    for key, value in results.items():
        optional = "Total" not in key
        metrics.add(key, value, optional)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# TODO: rename to paths.py or user.py?

import os
import xdg
import logging
from enum import Enum
from pathlib import Path


def get_config_dir():
    config_dir = os.path.join(xdg.xdg_config_home(), "mlonmcu")
    return config_dir


def init_config_dir():
    config_dir = Path(get_config_dir())
    config_dir.mkdir(parents=True, exist_ok=True)
    subdirs = ["environments", "models", "plugins"]
    files = ["environments.ini"]
    for subdir in subdirs:
        environments_dir = config_dir / subdir
        environments_dir.mkdir(exist_ok=True)
    for file in files:
        filepath = config_dir / file
        filepath.touch(exist_ok=True)


def get_environments_dir():
    environments_dir = os.path.join(get_config_dir(), "environments")
    return environments_dir


def get_environments_file():
    environments_file = os.path.join(get_config_dir(), "environments.ini")
    return environments_file


def get_plugins_dir():
    environments_dir = os.path.join(get_config_dir(), "plugins")
    return environments_dir


DEFAULTS = {
    "environment": "default",
    "template": "default",
}

env_subdirs = ["deps", "plugins"]


# class LogLevel(Enum):
#     DEBUG = 0
#     VERBOSE = 1
#     INFO = 2
#     WARNING = 3
#     ERROR = 4


class BaseConfig:
    def __repr__(self):
        return self.__class__.__name__ + "(" + str(vars(self)) + ")"


class DefaultsConfig(BaseConfig):
    # TODO: loglevels enum

    def __init__(
        self,
        log_level=logging.INFO,
        log_to_file=False,
        log_rotate=False,
        default_framework=None,
        default_backends={},
        default_target=None,
        cleanup_auto=False,
        cleanup_keep=100,
    ):
        self.log_level = log_level
        self.log_to_file = log_to_file
        self.log_rotate = log_rotate
        self.default_framework = default_framework
        self.default_backends = default_backends
        self.default_target = default_target
        self.cleanup_auto = cleanup_auto
        self.cleanup_keep = cleanup_keep


class PathConfig(BaseConfig):
    def __init__(self, path, base=None):
        if isinstance(path, str):
            self.path = Path(path)
        else:
            self.path = path
        if isinstance(base, str):
            self.base = Path(base)
        else:
            self.base = base
        if base:
            if not self.path.is_absolute():
                assert base is not None
                self.path = self.base / self.path
        # Resolve symlinks
        self.path = self.path.resolve()

    def __repr(self):
        return f"PathConfig({self.path})"


class RepoConfig(BaseConfig):
    def __init__(self, url, ref=None, options=None):
        self.url = url
        if ref is not None:
            assert isinstance(ref, str)
        self.ref = ref
        self.options = options if options is not None else {}
        assert isinstance(self.options, dict)

    @property
    def single_branch(self):
        value = self.options.get("single_branch", False)
        assert isinstance(value, bool)
        return value

    @property
    def recursive(self):
        value = self.options.get("recursive", True)
        assert isinstance(value, bool)
        return value

    @property
    def submodules(self):
        value = self.options.get("submodules", None)
        if value is not None:
            assert isinstance(value, list)
        return value


class BackendConfig(BaseConfig):
    def __init__(self, name, enabled=True, features={}):
        self.name = name
        self.enabled = enabled
        self.features = features


class FeatureKind(Enum):
    UNKNOWN = 0
    FRAMEWORK = 1
    BACKEND = 2
    TARGET = 3
    FRONTEND = 4


class FeatureConfig:
    def __init__(self, name, kind=FeatureKind.UNKNOWN, supported=True):
        self.name = name
        self.supported = supported

    def __repr__(self):
        return self.__class__.__name__ + "(" + str(vars(self)) + ")"


class FrameworkFeatureConfig(FeatureConfig):
    def __init__(self, name, framework, supported=True):
        super().__init__(name=name, kind=FeatureKind.FRONTEND, supported=supported)
        self.framework = framework


class BackendFeatureConfig(FeatureConfig):
    def __init__(self, name, backend, supported=True):
        super().__init__(name=name, kind=FeatureKind.FRONTEND, supported=supported)
        self.backend = backend


class PlatformFeatureConfig(FeatureConfig):
    def __init__(self, name, platform, supported=True):
        super().__init__(name=name, kind=FeatureKind.TARGET, supported=supported)
        self.platform = platform


class TargetFeatureConfig(FeatureConfig):
    def __init__(self, name, target, supported=True):
        super().__init__(name=name, kind=FeatureKind.TARGET, supported=supported)
        self.target = target


class FrontendFeatureConfig(FeatureConfig):
    def __init__(self, name, frontend, supported=True):
        super().__init__(name=name, kind=FeatureKind.FRONTEND, supported=supported)
        self.frontend = frontend


class FrameworkConfig(BaseConfig):
    def __init__(self, name, enabled=True, backends={}, features={}):
        self.name = name
        self.enabled = enabled
        self.backends = backends
        self.features = features


class FrontendConfig(BaseConfig):
    def __init__(self, name, enabled=True, features={}):
        self.name = name
        self.enabled = enabled
        self.features = features


class PlatformConfig(BaseConfig):
    def __init__(self, name, enabled=True, features={}):
        self.name = name
        self.enabled = enabled
        self.features = features


class TargetConfig(BaseConfig):
    def __init__(self, name, enabled=True, features={}):
        self.name = name
        self.enabled = enabled
        self.features = features



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def get_runtime_executor_tvmc_args(runtime, executor):
    return ["--runtime", runtime, "--executor", executor]


def get_pass_config_tvmc_args(pass_config):
    args = []
    for key, value in pass_config.items():
        args.extend(["--pass-config", f"{key}={value}"])
    return args


def get_disabled_pass_tvmc_args(disabled_passes):
    if len(disabled_passes) == 0:
        return []
    arg = ",".join(disabled_passes)
    return ["--disabled-pass", arg]


def get_input_shapes_tvmc_args(input_shapes):
    if input_shapes is None:
        return []
    arg = " ".join([f"{name}:[" + ",".join(list(map(str, dims))) + "]" for name, dims in input_shapes.items()])
    return ["--input-shapes", arg]


def check_allowed(target, name):
    common = ["libs", "model", "tag", "mcpu", "device", "keys"]
    if target == "c":
        return name in ["constants-byte-alignment", "workspace-bytes-alignment", "march"] + common
    elif target == "llvm":
        return (
            name
            in [
                "fast-math",
                "opt-level",
                "fast-math-ninf",
                "mattr",
                "num-cores",
                "fast-math-nsz",
                "fast-math-contract",
                "mtriple",
                "mfloat-abi",
                "fast-math-arcp",
                "fast-math-reassoc",
                "mabi",
                "num_cores",
            ]
            + common
        )

    else:
        return True


def gen_target_details_args(target, target_details):
    def helper(value):
        if isinstance(value, (bool, int)):
            # value = "true" if value else "false"
            value = str(int(value))
        return value

    return sum(
        [
            [f"--target-{target}-{key}", helper(value)]
            for key, value in target_details.items()
            if check_allowed(target, key)
        ],
        [],
    )


def gen_extra_target_details_args(extra_target_details):
    ret = []
    for extra_target, target_details in extra_target_details.items():
        if target_details:
            ret.append(gen_target_details_args(extra_target, target_details))
    return sum(ret, [])


def get_target_tvmc_args(target="c", extra_targets=[], target_details={}, extra_target_details={}):
    if extra_targets:
        assert isinstance(extra_targets, list)
    else:
        extra_targets = []
    if extra_target_details:
        assert isinstance(extra_target_details, dict)
    else:
        extra_target_details = {}

    return [
        "--target",
        ",".join(extra_targets + [target]),
        # TODO: provide a feature which sets these automatically depending on the chosen target
        *gen_target_details_args(target, target_details),
        *(gen_extra_target_details_args(extra_target_details)),
    ]


def get_tuning_records_tvmc_args(use_tuning_results, tuning_records_file):
    if use_tuning_results:
        assert tuning_records_file is not None, "No tuning records are available"
        return ["--tuning-records", str(tuning_records_file)]
    else:
        return []


def get_rpc_tvmc_args(enabled, key, hostname, port):
    return (
        [
            "--rpc-key",
            key,
            "--rpc-tracker",
            hostname + ":" + str(port),
        ]
        if enabled
        else []
    )


def get_tvmaot_tvmc_args(alignment_bytes, unpacked_api, runtime="crt", target="c", system_lib=False):
    ret = []
    if runtime == "crt":
        if unpacked_api:
            assert not system_lib, "Unpacked API is incompatible with system lib"
        ret += ["--runtime-crt-system-lib", str(int(system_lib))]

    ret += [
        *["--executor-aot-unpacked-api", str(int(unpacked_api))],
        *["--executor-aot-interface-api", "c" if unpacked_api else "packed"],
    ]
    if target == "c":
        ret += [
            *["--target-c-constants-byte-alignment", str(alignment_bytes)],
            *["--target-c-workspace-byte-alignment", str(alignment_bytes)],
        ]
    return ret


def get_tvmrt_tvmc_args(runtime="crt", system_lib=True, link_params=True):
    ret = []
    if runtime == "crt":
        ret += ["--runtime-crt-system-lib", str(int(system_lib))]
    ret += ["--executor-graph-link-params", str(int(link_params))]
    return ret


def get_data_tvmc_args(mode=None, ins_file=None, outs_file=None, print_top=10):
    ret = []
    if ins_file is not None:
        ret.extend(["--inputs", ins_file])
    else:
        if mode is not None:
            ret.extend(["--fill-mode", mode])

    if outs_file is not None:
        ret.extend(["--outputs", outs_file])

    if print_top is not None and print_top > 0:
        ret.extend(["--print-top", print_top])

    return ret


def get_bench_tvmc_args(print_time=False, profile=False, end_to_end=False, repeat=1, number=1):
    ret = []

    if print_time:
        ret.append("--print-time")

    if profile:
        ret.append("--profile")

    if end_to_end:
        ret.append("--end-to-end")

    if repeat:
        ret.extend(["--repeat", str(repeat)])

    if number:
        ret.extend(["--number", str(number)])

    return ret


def get_desired_layout_args(layouts, ops, mapping):
    if mapping:
        assert layouts is None, "desired_layout not allowed when using desired_layouts_map"
        assert ops is None, "desired_layout_ops not allowed when using desired_layouts_map"
        layouts = mapping.values()
        ops = mapping.keys()

    if layouts is None:
        layouts = []

    if ops is None:
        ops = []

    if layouts and ops:
        assert len(layouts) == len(ops) or len(layouts) == 1

    ret = []
    if layouts:
        ret.extend(["--desired-layout", *layouts])

    if ops:
        ret.extend(["--desired-layout-ops", *ops])

    return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Console script for mlonmcu."""

import os
import argparse
import sys
import subprocess
import platform


from mlonmcu.logging import get_logger
import mlonmcu.cli.init as init
import mlonmcu.cli.setup as setup
import mlonmcu.cli.flow as flow
import mlonmcu.cli.cleanup as cleanup
import mlonmcu.cli.export as export
import mlonmcu.cli.env as env
import mlonmcu.cli.models as models
from .common import handle_logging_flags, add_common_options
from ..version import __version__

logger = get_logger()


def handle_docker(args):
    if args.docker:
        home = os.environ.get("MLONMCU_HOME")
        assert home is not None, (
            "To use the --docker functionality, please export the MLONMCU_HOME environment variable"
            + " to a directory which should be mounted by the container"
        )
        exec_args = sys.argv[1:]
        exec_args.remove("--docker")
        docker = subprocess.Popen(
            [
                "docker-compose",
                "-f",
                "docker/docker-compose.yml",
                "run",
                "-e",
                f"MLONMCU_HOME={home}",
                "--rm",
                "mlonmcu",
                "python3",
                "-m",
                "mlonmcu.cli.main",
                *exec_args,
            ],
            env={"MLONMCU_HOME": home},
        )
        stdout, stderr = docker.communicate()
        exit_code = docker.wait()
        if exit_code > 0:
            logger.warning(f"Docker compose process completed with exit code: {exit_code}")
        sys.exit(exit_code)

    if platform.system() in ["Darwin", "Windows"]:
        raise RuntimeError(
            "Only Linux is supported at the Moment. If you have Docker installed, you may want to"
            + " try running this script using the `--docker` flag."
        )


# def main(args):
def main(args=None):
    """Console script for mlonmcu."""

    parser = argparse.ArgumentParser(
        description="ML on MCU Flow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # parser.add_argument('_', nargs='*')
    parser.add_argument("-V", "--version", action="version", version="mlonmcu " + __version__)
    add_common_options(parser)
    subparsers = parser.add_subparsers(dest="subcommand")  # this line changed
    init.get_parser(subparsers)
    setup.get_parser(subparsers)
    flow.get_parser(subparsers)
    cleanup.get_parser(subparsers)
    export.get_parser(subparsers)
    env.get_parser(subparsers)
    models.get_parser(subparsers)
    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()
    handle_logging_flags(args)
    handle_docker(args)

    if hasattr(args, "func"):
        args.func(args)
    else:
        print("Invalid subcommand for `mlonmcu`!")
        parser.print_help(sys.stderr)

    return 0


if __name__ == "__main__":
    sys.exit(main(args=sys.argv[1:]))  # pragma: no cover



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from typing import Tuple

# import json

from .backend import TVMBackend
from .wrapper import generate_tvmrt_wrapper, generate_wrapper_header
from mlonmcu.flow.backend import main
from mlonmcu.config import str2bool
from mlonmcu.artifact import Artifact, ArtifactFormat, lookup_artifacts
from .tvmc_utils import get_tvmrt_tvmc_args
from .model_info import get_relay_model_info


class TVMRTBackend(TVMBackend):
    FEATURES = TVMBackend.FEATURES | {"debug_arena"}

    DEFAULTS = {
        **TVMBackend.DEFAULTS,
        "debug_arena": False,
        "arena_size": 2**20,  # Can not be detemined automatically (Very large)
        # TODO: arena size warning!
    }

    name = "tvmrt"

    def __init__(self, runtime="crt", fmt="mlf", system_lib=True, features=None, config=None):
        super().__init__(
            executor="graph", runtime=runtime, fmt=fmt, system_lib=system_lib, features=features, config=config
        )

    @property
    def arena_size(self):
        size = self.config["arena_size"]
        return int(size) if size else None

    @property
    def debug_arena(self):
        value = self.config["debug_arena"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    def get_tvmc_compile_args(self, out, dump=None):
        return super().get_tvmc_compile_args(out, dump=dump) + get_tvmrt_tvmc_args(system_lib=self.system_lib)

    def generate(self) -> Tuple[dict, dict]:
        artifacts, metrics = super().generate()
        assert len(artifacts) == 1 and "default" in artifacts
        artifacts = artifacts["default"]
        assert len(metrics) == 1 and "default" in metrics
        metrics = metrics["default"]
        if self.generate_wrapper:
            workspace_size = self.arena_size
            assert workspace_size >= 0
            graph_artifact = lookup_artifacts(artifacts, f"{self.prefix}.graph")[0]
            graph = graph_artifact.content
            params_artifact = lookup_artifacts(artifacts, f"{self.prefix}.params")[0]
            params = params_artifact.raw
            if (not self.model_info) or self.refresh_model_info:
                try:
                    relay_artifact = lookup_artifacts(artifacts, f"{self.prefix}.relay")[0]
                    self.model_info = get_relay_model_info(relay_artifact.content)
                except Exception:
                    assert self.model_info is not None, "Model info missing"
            wrapper_src = generate_tvmrt_wrapper(
                graph, params, self.model_info, workspace_size, debug_arena=self.debug_arena
            )
            artifacts.append(Artifact("rt_wrapper.c", content=wrapper_src, fmt=ArtifactFormat.SOURCE))
            header_src = generate_wrapper_header()
            artifacts.append(Artifact("tvm_wrapper.h", content=header_src, fmt=ArtifactFormat.SOURCE))
        metrics.add("Workspace Size [B]", workspace_size, True)
        return {"default": artifacts}, {"default": metrics}


if __name__ == "__main__":
    sys.exit(
        main(
            TVMRTBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover
